"""
Container-type object and associated functions

    :copyright: 2021 by The Autoprotocol Development Team, see AUTHORS
        for more details.
    :license: BSD, see LICENSE for more details

"""
import re

from collections import namedtuple

from .container import Well
from .unit import Unit


class ContainerType(
    namedtuple(
        "ContainerType",
        [
            "name",
            "is_tube",
            "well_count",
            "well_depth_mm",
            "well_volume_ul",
            "well_coating",
            "sterile",
            "cover_types",
            "seal_types",
            "capabilities",
            "shortname",
            "col_count",
            "dead_volume_ul",
            "safe_min_volume_ul",
            "true_max_vol_ul",
            "vendor",
            "cat_no",
            "prioritize_seal_or_cover",
        ],
    )
):
    """
    The ContainerType class holds the capabilities and properties of a
    particular container type.

    Parameters
    ----------
    name : str
      Full name describing a ContainerType.
    is_tube : bool
      Indicates whether a ContainerType is a tube (container with one well).
    well_count : int
      Number of wells a ContainerType contains.
    well_depth_mm : float
      Depth of well(s) contained in a ContainerType in millimeters.
    well_volume_ul : Unit
      Maximum volume of well(s) contained in a ContainerType in microliters.
    well_coating : str
      Coating of well(s) in container (ex. collagen).
    sterile : bool
      Indicates whether a ContainerType is sterile.
    cover_types: list
        List of valid covers associated with a ContainerType.
    seal_types: list
        List of valid seals associated with a ContainerType.
    capabilities : list
      List of capabilities associated with a ContainerType (ex. ["spin",
        "incubate"]).
    shortname : str
      Short name used to refer to a ContainerType.
    col_count : int
      Number of columns a ContainerType contains.
    dead_volume_ul : Unit
      Volume of liquid that cannot be aspirated from any given well of a
      ContainerType via liquid-handling.
    safe_min_volume_ul : Unit
      Minimum volume of liquid to ensure adequate volume for liquid-handling
      aspiration from any given well of a ContainerType.
    true_max_vol_ul : Unit, optional
      Maximum volume of well(s) in microliters, often same value as
      well_volume_ul (maximum working volume), however, some ContainerType(s)
      can have a different value corresponding to a true maximum volume
      of a well (ex. echo compatible containers)
    vendor: str, optional
      ContainerType commercial vendor, if available.
    cat_no: str, optional
      ContainerType vendor catalog number, if available.
    prioritize_seal_or_cover: str, optional
        "seal" or "cover", determines whether to prioritize sealing or covering
        defaults to "seal"
    """

    def __new__(
        cls,
        name,
        is_tube,
        well_count,
        well_depth_mm,
        well_volume_ul,
        well_coating,
        sterile,
        cover_types,
        seal_types,
        capabilities,
        shortname,
        col_count,
        dead_volume_ul,
        safe_min_volume_ul,
        true_max_vol_ul=None,
        vendor=None,
        cat_no=None,
        prioritize_seal_or_cover="seal",
    ):
        true_max_vol_ul = true_max_vol_ul or well_volume_ul
        assert true_max_vol_ul >= well_volume_ul, (
            f"{name} does not contain valid true_max_vol_ul: "
            f"{true_max_vol_ul} and well_volume_ul {well_volume_ul}"
        )
        return super(ContainerType, cls).__new__(
            cls,
            name,
            is_tube,
            well_count,
            well_depth_mm,
            well_volume_ul,
            well_coating,
            sterile,
            cover_types,
            seal_types,
            capabilities,
            shortname,
            col_count,
            dead_volume_ul,
            safe_min_volume_ul,
            true_max_vol_ul,
            vendor,
            cat_no,
            prioritize_seal_or_cover,
        )

    @staticmethod
    def well_from_coordinates_static(row, row_count, col, col_count):
        if row >= row_count:
            raise ValueError(
                f"0-indexed row {row} is outside of the bounds of {row_count}"
            )

        if col >= col_count:
            raise ValueError(
                f"0-indexed column {col} is outside of the bounds of {col_count}"
            )

        return row * col_count + col

    def well_from_coordinates(self, row, column):
        """
        Gets the well at 0-indexed position (row, column) within the container.
        The origin is in the top left corner.

        Parameters
        ----------
        row : int
            The 0-indexed row index of the well to be fetched
        column : int
            The 0-indexed column index of the well to be fetched

        Returns
        -------
        Int
            The robotized index of the well at at position (row, column)

        Raises
        ------
        ValueError
            if the specified row is outside the bounds of the container_type
        ValueError
            if the specified column is outside the bounds of the container_type
        """
        return ContainerType.well_from_coordinates_static(
            row, self.row_count(), column, self.col_count
        )

    @staticmethod
    def robotize_static(well_ref, well_count, col_count):
        if isinstance(well_ref, list):
            return [
                ContainerType.robotize_static(well, well_count, col_count)
                for well in well_ref
            ]

        if not isinstance(well_ref, (str, int, Well)):
            raise TypeError(
                f"ContainerType.robotize(): Well reference "
                f"({well_ref}) given is not of type 'str', 'int', "
                f"or 'Well'."
            )

        if isinstance(well_ref, Well):
            well_ref = well_ref.index
        well_ref = str(well_ref)
        m = re.match(r"([a-z])([a-z]?)(\d+)$", well_ref, re.I)
        if m:
            row = ord(m.group(1).upper()) - ord("A")
            if m.group(2):
                row = 26 * (row + 1) + ord(m.group(2).upper()) - ord("A")
            col = int(m.group(3)) - 1
            row_count = well_count // col_count
            return ContainerType.well_from_coordinates_static(
                row, row_count, col, col_count
            )
        else:
            m = re.match(r"\d+$", well_ref)
            if m:
                well_num = int(m.group(0))
                # Check bounds
                if well_num >= well_count or well_num < 0:
                    raise ValueError(
                        "ContainerType.robotize(): Well number "
                        "given exceeds container dimensions."
                    )
                return well_num
            else:
                raise ValueError(
                    "ContainerType.robotize(): Well must be in "
                    "'A1' format or be an integer."
                )

    def robotize(self, well_ref):
        """
        Return a robot-friendly well reference from a number of well reference
        formats.

        Example Usage:

        .. code-block:: python

            >>> p = Protocol()
            >>> my_plate = p.ref("my_plate", cont_type="6-flat", discard=True)
            >>> my_plate.robotize("A1")
            0
            >>> my_plate.robotize("5")
            5
            >>> my_plate.robotize(my_plate.well(3))
            3
            >>> my_plate.robotize(["A1", "A2"])
            [0, 1]

        Parameters
        ----------
        well_ref : str, int, Well, list[str or int or Well]
            Well reference to be robotized in string, integer or Well object
            form. Also accepts lists of str, int or Well.

        Returns
        -------
        int or list
            Single or list of Well references passed as row-wise integer
            (left-to-right, top-to-bottom, starting at 0 = A1).

        Raises
        ------
        TypeError
            If well reference given is not an accepted type.
        ValueError
            If well reference given exceeds container dimensions.
        ValueError
            If well reference given is in an invalid format.
        """
        return ContainerType.robotize_static(well_ref, self.well_count, self.col_count)

    @staticmethod
    def humanize_static(well_ref, well_count, col_count):
        ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        if isinstance(well_ref, list):
            return [
                ContainerType.humanize_static(well, well_count, col_count)
                for well in well_ref
            ]

        if not isinstance(well_ref, (int, str)):
            raise TypeError(
                "ContainerType.humanize(): Well reference given "
                "is not of type 'int' or 'str'."
            )
        try:
            well_ref = int(well_ref)
        except:
            raise TypeError(
                "ContainerType.humanize(): Well reference given"
                "is not parseable into 'int' format."
            )
        # Check bounds
        if well_ref >= well_count or well_ref < 0:
            raise ValueError(
                "ContainerType.humanize(): Well reference "
                "given exceeds container dimensions."
            )
        idx = ContainerType.robotize_static(well_ref, well_count, col_count)
        row, col = (idx // col_count, idx % col_count)
        if row >= len(ALPHABET):
            return ALPHABET[row // 26 - 1] + ALPHABET[row % 26] + str(col + 1)
        else:
            return ALPHABET[row] + str(col + 1)

    def humanize(self, well_ref):
        """
        Return the human readable form of a well index based on the well
        format of this ContainerType.

        Example Usage:

        .. code-block:: python

            >>> p = Protocol()
            >>> my_plate = p.ref("my_plate", cont_type="6-flat", discard=True)
            >>> my_plate.humanize(0)
            'A1'
            >>> my_plate.humanize(5)
            'B3'
            >>> my_plate.humanize('0')
            'A1'

        Parameters
        ----------
        well_ref : int, str, list[int or str]
            Well reference to be humanized in integer or string form.
            If string is provided, it has to be parseable into an int.
            Also accepts lists of int or str

        Returns
        -------
        well_ref : str
            Well index passed as human-readable form.

        Raises
        ------
        TypeError
            If well reference given is not an accepted type.
        ValueError
            If well reference given exceeds container dimensions.

        """
        return ContainerType.humanize_static(well_ref, self.well_count, self.col_count)

    def decompose(self, idx):
        """
        Return the (col, row) corresponding to the given well index.

        Parameters
        ----------
        idx : str or int
            Well index in either human-readable or integer form.

        Returns
        -------
        tuple
            tuple containing the column number and row number of the given
            well_ref.

        Raises
        ------
        TypeError
            Index given is not of the right parameter type

        """
        if not isinstance(idx, (int, str, Well)):
            raise TypeError("Well index given is not of type 'int' or " "'str'.")
        idx = self.robotize(idx)
        return (idx // self.col_count, idx % self.col_count)

    def row_count(self):
        """
        Return the number of rows of this ContainerType.

        """
        return self.well_count // self.col_count


#:
FLAT384 = ContainerType(
    name="384-well UV flat-bottom plate",
    well_count=384,
    well_depth_mm=None,
    well_volume_ul=Unit(90.0, "microliter"),
    well_coating=None,
    sterile=False,
    is_tube=False,
    cover_types=["standard", "universal"],
    seal_types=["ultra-clear", "foil"],
    prioritize_seal_or_cover="cover",
    capabilities=[
        "liquid_handle",
        "spin",
        "absorbance",
        "fluorescence",
        "luminescence",
        "incubate",
        "gel_separate",
        "gel_purify",
        "cover",
        "dispense",
        "seal",
    ],
    shortname="384-flat",
    col_count=24,
    dead_volume_ul=Unit(7, "microliter"),
    safe_min_volume_ul=Unit(15, "microliter"),
    vendor="Corning",
    cat_no="3706",
)


#:
PCR384 = ContainerType(
    name="384-well PCR plate",
    well_count=384,
    well_depth_mm=None,
    well_volume_ul=Unit(40.0, "microliter"),
    well_coating=None,
    sterile=None,
    is_tube=False,
    cover_types=None,
    seal_types=["ultra-clear", "foil"],
    capabilities=[
        "liquid_handle",
        "spin",
        "thermocycle",
        "incubate",
        "gel_separate",
        "gel_purify",
        "seal",
        "dispense",
    ],
    shortname="384-pcr",
    col_count=24,
    dead_volume_ul=Unit(2, "microliter"),
    safe_min_volume_ul=Unit(3, "microliter"),
    vendor="Eppendorf",
    cat_no="951020539",
)

#:
ECHO384 = ContainerType(
    # Compatible with Labcyte Echo 520, 525 and 55x
    name="384-well Echo plate",
    well_count=384,
    well_depth_mm=None,
    well_volume_ul=Unit(65.0, "microliter"),
    well_coating=None,
    sterile=None,
    is_tube=False,
    cover_types=["universal"],
    seal_types=["foil", "ultra-clear"],
    capabilities=["liquid_handle", "seal", "spin", "incubate", "dispense", "cover"],
    shortname="384-echo",
    col_count=24,
    dead_volume_ul=Unit(15, "microliter"),
    safe_min_volume_ul=Unit(15, "microliter"),
    true_max_vol_ul=Unit(135, "microliter"),
    vendor="Labcyte",
    cat_no="PP-0200",
)

#:
FLAT384WHITELV = ContainerType(
    name="384-well flat-bottom low volume plate",
    well_count=384,
    well_depth_mm=9.39,
    well_volume_ul=Unit(40.0, "microliter"),
    well_coating=None,
    sterile=False,
    is_tube=False,
    cover_types=["standard", "universal"],
    seal_types=None,
    capabilities=[
        "absorbance",
        "cover",
        "dispense",
        "fluorescence",
        "image_plate",
        "incubate",
        "luminescence",
        "liquid_handle",
        "spin",
        "uncover",
    ],
    shortname="384-flat-white-white-lv",
    col_count=24,
    dead_volume_ul=Unit(5, "microliter"),
    safe_min_volume_ul=Unit(15, "microliter"),
    vendor="Corning",
    cat_no="3824",
)

#:
FLAT384WHITETC = ContainerType(
    name="384-well flat-bottom low flange plate",
    well_count=384,
    well_depth_mm=11.43,
    well_volume_ul=Unit(80.0, "microliter"),
    well_coating=None,
    sterile=True,
    is_tube=False,
    cover_types=["standard", "universal"],
    seal_types=None,
    capabilities=[
        "absorbance",
        "cover",
        "dispense",
        "fluorescence",
        "image_plate",
        "incubate",
        "luminescence",
        "liquid_handle",
        "spin",
        "uncover",
    ],
    shortname="384-flat-white-white-tc",
    col_count=24,
    dead_volume_ul=Unit(20, "microliter"),
    safe_min_volume_ul=Unit(30, "microliter"),
    vendor="Corning",
    cat_no="3570",
)

#:
FLAT384CLEAR = ContainerType(
    name="384-well fully clear high binding plate",
    well_count=384,
    well_depth_mm=11.43,
    well_volume_ul=Unit(80.0, "microliter"),
    well_coating="high bind",
    sterile=False,
    is_tube=False,
    cover_types=["standard", "universal", "low_evaporation"],
    seal_types=["ultra-clear", "foil"],
    capabilities=[
        "incubate",
        "seal",
        "image_plate",
        "dispense",
        "spin",
        "absorbance",
        "cover",
        "fluorescence",
        "luminescence",
        "liquid_handle",
        "uncover",
    ],
    shortname="384-flat-clear-clear",
    col_count=24,
    dead_volume_ul=Unit(5, "microliter"),
    safe_min_volume_ul=Unit(20, "microliter"),
    vendor="Corning",
    cat_no="3700",
)

#:
FLAT96 = ContainerType(
    name="96-well flat-bottom plate",
    well_count=96,
    well_depth_mm=None,
    well_volume_ul=Unit(340.0, "microliter"),
    well_coating=None,
    sterile=False,
    is_tube=False,
    cover_types=["low_evaporation", "standard", "universal"],
    seal_types=None,
    capabilities=[
        "liquid_handle",
        "spin",
        "absorbance",
        "fluorescence",
        "luminescence",
        "incubate",
        "gel_separate",
        "gel_purify",
        "cover",
        "dispense",
    ],
    shortname="96-flat",
    col_count=12,
    dead_volume_ul=Unit(25, "microliter"),
    safe_min_volume_ul=Unit(65, "microliter"),
    vendor="Corning",
    cat_no="3632",
)

#:
FLAT96UV = ContainerType(
    name="96-well flat-bottom UV transparent plate",
    well_count=96,
    well_depth_mm=None,
    well_volume_ul=Unit(340.0, "microliter"),
    well_coating=None,
    sterile=False,
    is_tube=False,
    cover_types=["low_evaporation", "standard", "universal"],
    seal_types=None,
    capabilities=[
        "liquid_handle",
        "spin",
        "absorbance",
        "fluorescence",
        "luminescence",
        "incubate",
        "gel_separate",
        "gel_purify",
        "cover",
        "dispense",
    ],
    shortname="96-flat-uv",
    col_count=12,
    dead_volume_ul=Unit(25, "microliter"),
    safe_min_volume_ul=Unit(65, "microliter"),
    vendor="Corning",
    cat_no="3635",
)

#:
PCR96 = ContainerType(
    name="96-well PCR plate",
    well_count=96,
    well_depth_mm=None,
    well_volume_ul=Unit(160.0, "microliter"),
    well_coating=None,
    sterile=None,
    is_tube=False,
    cover_types=None,
    seal_types=["ultra-clear", "foil"],
    capabilities=[
        "liquid_handle",
        "sangerseq",
        "spin",
        "thermocycle",
        "incubate",
        "gel_separate",
        "gel_purify",
        "seal",
        "dispense",
    ],
    shortname="96-pcr",
    col_count=12,
    dead_volume_ul=Unit(3, "microliter"),
    safe_min_volume_ul=Unit(5, "microliter"),
    vendor="Eppendorf",
    cat_no="951020619",
)

#:
DEEP96 = ContainerType(
    name="96-well extended capacity plate",
    well_count=96,
    well_depth_mm=None,
    well_volume_ul=Unit(2000.0, "microliter"),
    well_coating=None,
    sterile=False,
    cover_types=["standard", "universal"],
    seal_types=["breathable"],
    prioritize_seal_or_cover="cover",
    capabilities=[
        "liquid_handle",
        "incubate",
        "gel_separate",
        "gel_purify",
        "cover",
        "dispense",
        "seal",
    ],
    shortname="96-deep",
    is_tube=False,
    col_count=12,
    dead_volume_ul=Unit(5, "microliter"),
    safe_min_volume_ul=Unit(30, "microliter"),
    vendor="Corning",
    cat_no="3961",
)

#:
V96KF = ContainerType(
    name="96-well v-bottom King Fisher plate",
    well_count=96,
    well_depth_mm=None,
    well_volume_ul=Unit(200.0, "microliter"),
    well_coating=None,
    sterile=False,
    cover_types=["standard"],
    seal_types=None,
    capabilities=[
        "liquid_handle",
        "incubate",
        "gel_separate",
        "mag_dry",
        "mag_incubate",
        "mag_collect",
        "mag_release",
        "mag_mix",
        "cover",
        "dispense",
    ],
    shortname="96-v-kf",
    is_tube=False,
    col_count=12,
    dead_volume_ul=Unit(20, "microliter"),
    safe_min_volume_ul=Unit(20, "microliter"),
    vendor="Fisher",
    cat_no="22-387-030",
)

#:
DEEP96KF = ContainerType(
    name="96-well extended capacity King Fisher plate",
    well_count=96,
    well_depth_mm=None,
    well_volume_ul=Unit(1000.0, "microliter"),
    well_coating=None,
    sterile=False,
    cover_types=["standard"],
    seal_types=None,
    capabilities=[
        "liquid_handle",
        "incubate",
        "gel_separate",
        "mag_dry",
        "mag_incubate",
        "mag_collect",
        "mag_release",
        "mag_mix",
        "cover",
        "dispense",
    ],
    shortname="96-deep-kf",
    is_tube=False,
    col_count=12,
    dead_volume_ul=Unit(50, "microliter"),
    safe_min_volume_ul=Unit(50, "microliter"),
    vendor="Fisher",
    cat_no="22-387-031",
)

#:
V96CC = ContainerType(
    name="96-well cell culture multiple well plate, V bottom",
    well_count=96,
    well_depth_mm=10.668,
    well_volume_ul=Unit(200.0, "microliter"),
    well_coating=None,
    sterile=True,
    is_tube=False,
    cover_types=["standard", "universal", "low_evaporation", "ultra-clear", "foil"],
    seal_types=None,
    capabilities=[
        "dispense",
        "spin",
        "seal",
        "unseal",
        "liquid_handle",
        "cover",
        "echo_dest",
        "spectrophotometry",
        "image_plate",
        "incubate",
        "uncover",
        "dispense-destination",
        "envision",
        "absorbance",
        "fluorescence",
    ],
    shortname="96-well-v-bottom",
    col_count=12,
    dead_volume_ul=Unit(75.0, "microliter"),
    safe_min_volume_ul=Unit(5.0, "microliter"),
    true_max_vol_ul=Unit(320.0, "microliter"),
    vendor="Corning",
    cat_no="3894",
)

#:
DEEP24 = ContainerType(
    name="24-well extended capacity plate",
    well_count=24,
    well_depth_mm=None,
    well_volume_ul=Unit(10000.0, "microliter"),
    well_coating=None,
    sterile=False,
    cover_types=None,
    seal_types=["foil", "breathable"],
    capabilities=[
        "liquid_handle",
        "incubate",
        "gel_separate",
        "gel_purify",
        "dispense",
        "seal",
    ],
    shortname="24-deep",
    is_tube=False,
    col_count=6,
    dead_volume_ul=Unit(15, "microliter"),
    safe_min_volume_ul=Unit(60, "microliter"),
    vendor="E&K Scientific",
    cat_no="EK-2053-S",
)

#:
MICRO2 = ContainerType(
    name="2mL Microcentrifuge tube",
    well_count=1,
    well_depth_mm=None,
    well_volume_ul=Unit(2000.0, "microliter"),
    well_coating=None,
    sterile=False,
    cover_types=None,
    seal_types=None,
    capabilities=["liquid_handle", "gel_separate", "gel_purify", "incubate", "spin"],
    shortname="micro-2.0",
    is_tube=True,
    col_count=1,
    dead_volume_ul=Unit(5, "microliter"),
    safe_min_volume_ul=Unit(40, "microliter"),
    vendor="E&K Scientific",
    cat_no="280200",
)

#:
MICRO15 = ContainerType(
    name="1.5mL Microcentrifuge tube",
    well_count=1,
    well_depth_mm=None,
    well_volume_ul=Unit(1500.0, "microliter"),
    well_coating=None,
    sterile=False,
    cover_types=None,
    seal_types=None,
    capabilities=["liquid_handle", "gel_separate", "gel_purify", "incubate", "spin"],
    shortname="micro-1.5",
    is_tube=True,
    col_count=1,
    dead_volume_ul=Unit(20, "microliter"),
    safe_min_volume_ul=Unit(20, "microliter"),
    vendor="USA Scientific",
    cat_no="1615-5500",
)

#:
FLAT6 = ContainerType(
    name="6-well cell culture plate",
    well_count=6,
    well_depth_mm=None,
    well_volume_ul=Unit(5000.0, "microliter"),
    well_coating=None,
    sterile=False,
    cover_types=["standard", "universal"],
    seal_types=None,
    capabilities=["cover", "incubate", "image_plate"],
    shortname="6-flat",
    is_tube=False,
    col_count=3,
    dead_volume_ul=Unit(400, "microliter"),
    safe_min_volume_ul=Unit(600, "microliter"),
    vendor="Eppendorf",
    cat_no="30720016",
)

#:
FLAT1 = ContainerType(
    name="1-well flat-bottom plate",
    well_count=1,
    well_depth_mm=None,
    well_volume_ul=Unit(80000.0, "microliter"),
    well_coating=None,
    sterile=False,
    cover_types=["universal"],
    seal_types=None,
    capabilities=["cover", "incubate"],
    shortname="1-flat",
    is_tube=False,
    col_count=1,
    dead_volume_ul=Unit(36000, "microliter"),
    safe_min_volume_ul=Unit(40000, "microliter"),
    vendor="Fisher",
    cat_no="267060",
)

#:
FLAT6TC = ContainerType(
    name="6-well TC treated plate",
    well_count=6,
    well_depth_mm=None,
    well_volume_ul=Unit(5000.0, "microliter"),
    well_coating=None,
    sterile=False,
    cover_types=["standard", "universal"],
    seal_types=None,
    capabilities=["cover", "incubate", "image_plate"],
    shortname="6-flat-tc",
    is_tube=False,
    col_count=3,
    dead_volume_ul=Unit(400, "microliter"),
    safe_min_volume_ul=Unit(600, "microliter"),
    vendor="Eppendorf",
    cat_no="30720113",
)

#:
RESSW96HP = ContainerType(
    name="96-well singlewell highprofile reservoir",
    well_count=1,
    well_depth_mm=None,
    well_volume_ul=Unit(170.0, "milliliter"),
    well_coating=None,
    sterile=False,
    is_tube=False,
    cover_types=["universal"],
    seal_types=None,
    capabilities=["liquid_handle", "incubate", "cover", "dispense", "liquid_handle"],
    shortname="res-sw96-hp",
    col_count=1,
    dead_volume_ul=Unit(25, "milliliter"),
    safe_min_volume_ul=Unit(30, "milliliter"),
    true_max_vol_ul=Unit(280.0, "milliliter"),
    vendor="Axygen",
    cat_no="res-sw96-hp",
)

#:
RESMW8HP = ContainerType(
    name="8-row multiwell highprofile reservoir",
    well_count=8,
    well_depth_mm=None,
    well_volume_ul=Unit(24.0, "milliliter"),
    well_coating=None,
    sterile=False,
    is_tube=False,
    cover_types=["universal"],
    seal_types=None,
    capabilities=["liquid_handle", "incubate", "cover", "dispense", "liquid_handle"],
    shortname="res-mw8-hp",
    col_count=1,
    dead_volume_ul=Unit(2.5, "milliliter"),
    safe_min_volume_ul=Unit(5, "milliliter"),
    true_max_vol_ul=Unit(32.0, "milliliter"),
    vendor="Axygen",
    cat_no="res-mw8-hp",
)

#:
RESMW12HP = ContainerType(
    name="12-column multiwell highprofile reservoir",
    well_count=12,
    well_depth_mm=None,
    well_volume_ul=Unit(15.75, "milliliter"),
    well_coating=None,
    sterile=False,
    is_tube=False,
    cover_types=["universal"],
    seal_types=None,
    capabilities=["liquid_handle", "incubate", "cover", "dispense", "liquid_handle"],
    shortname="res-mw12-hp",
    col_count=12,
    dead_volume_ul=Unit(1.8, "milliliter"),
    safe_min_volume_ul=Unit(5, "milliliter"),
    true_max_vol_ul=Unit(21.0, "milliliter"),
    vendor="Axygen",
    cat_no="res-mw12-hp",
)

#:
FLAT96CLEARTC = ContainerType(
    name="96-well flat-bottom TC treated plate",
    well_count=96,
    well_depth_mm=None,
    well_volume_ul=Unit(340.0, "microliter"),
    well_coating=None,
    sterile=False,
    is_tube=False,
    cover_types=["low_evaporation", "standard", "universal"],
    seal_types=None,
    capabilities=[
        "liquid_handle",
        "spin",
        "absorbance",
        "fluorescence",
        "luminescence",
        "incubate",
        "gel_separate",
        "gel_purify",
        "cover",
        "dispense",
    ],
    shortname="96-flat-clear-clear-tc",
    col_count=12,
    dead_volume_ul=Unit(25, "microliter"),
    safe_min_volume_ul=Unit(65, "microliter"),
    vendor="Eppendorf",
    cat_no="0030730119",
)

#:
V384CLEAR = ContainerType(
    name="384-well v-bottom polypropylene plate",
    well_count=384,
    well_depth_mm=None,
    well_volume_ul=Unit(120.0, "microliter"),
    well_coating=None,
    sterile=False,
    is_tube=False,
    cover_types=["universal", "standard"],
    seal_types=["ultra-clear", "foil"],
    capabilities=[
        "incubate",
        "seal",
        "image_plate",
        "liquid_handle",
        "dispense",
        "spin",
        "mag_dry",
        "mag_incubate",
        "mag_collect",
        "mag_release",
        "mag_mix",
        "absorbance",
        "fluorescence",
        "luminescence",
        "cover",
        "thermocycle",
    ],
    shortname="384-v-clear-clear",
    col_count=24,
    dead_volume_ul=Unit(13, "microliter"),
    safe_min_volume_ul=Unit(18, "microliter"),
    vendor="Greiner",
    cat_no="781280",
)

#:
ROUND384CLEAR = ContainerType(
    name="384-well round-bottom plate",
    well_count=384,
    well_depth_mm=None,
    well_volume_ul=Unit(70.0, "microliter"),
    well_coating=None,
    sterile=False,
    is_tube=False,
    cover_types=["universal", "standard"],
    seal_types=["ultra-clear", "foil"],
    capabilities=[
        "incubate",
        "seal",
        "image_plate",
        "liquid_handle",
        "dispense",
        "spin",
        "mag_dry",
        "mag_incubate",
        "mag_collect",
        "mag_release",
        "mag_mix",
        "absorbance",
        "fluorescence",
        "luminescence",
        "cover",
        "thermocycle",
    ],
    shortname="384-round-clear-clear",
    col_count=24,
    dead_volume_ul=Unit(15, "microliter"),
    safe_min_volume_ul=Unit(20, "microliter"),
    vendor="Corning",
    cat_no="3657",
)

#:
RESSW384LP = ContainerType(
    name="384-well singlewell lowprofile reservoir",
    well_count=1,
    well_depth_mm=None,
    well_volume_ul=Unit(35, "milliliter"),
    well_coating=None,
    sterile=False,
    is_tube=False,
    cover_types=["universal"],
    seal_types=None,
    capabilities=[
        "liquid_handle",
        "incubate",
        "cover",
        "dispense",
        "liquid_handle",
        "sbs384_compatible",
    ],
    shortname="res-sw384-lp",
    col_count=1,
    dead_volume_ul=Unit(10, "milliliter"),
    safe_min_volume_ul=Unit(30, "milliliter"),
    true_max_vol_ul=Unit(92, "milliliter"),
    vendor="Axygen",
    cat_no="res-sw384-lp",
)

#:
ECHO384LDV = ContainerType(
    # Compatible with Labcyte Echo 520 and 55x only
    name="384-well Echo low dead volume plate",
    well_count=384,
    well_depth_mm=None,
    well_volume_ul=Unit(14.0, "microliter"),
    well_coating=None,
    sterile=None,
    is_tube=False,
    cover_types=["universal"],
    seal_types=["foil", "ultra-clear"],
    capabilities=["liquid_handle", "seal", "spin", "incubate", "dispense", "cover"],
    shortname="384-echo-ldv",
    col_count=24,
    dead_volume_ul=Unit(2.5, "microliter"),
    safe_min_volume_ul=Unit(2.5, "microliter"),
    vendor="Labcyte",
    cat_no="LP-0200",
)

#:
ECHO384LDVPLUS = ContainerType(
    # Compatible with Labcyte Echo 525 only
    name="384-well Echo low dead volume plus plate",
    well_count=384,
    well_depth_mm=None,
    well_volume_ul=Unit(14.0, "microliter"),
    well_coating="PureCoat Amine Surface",
    sterile=None,
    is_tube=False,
    cover_types=["universal"],
    seal_types=["foil", "ultra-clear"],
    capabilities=["liquid_handle", "seal", "spin", "incubate", "dispense", "cover"],
    shortname="384-echo-ldv-plus",
    col_count=24,
    dead_volume_ul=Unit(4.5, "microliter"),
    safe_min_volume_ul=Unit(4.5, "microliter"),
    vendor="Labcyte",
    cat_no="LPL-0200",
)

#:
FLAT384WHITECLEAR = ContainerType(
    name="384-well flat-bottom polystyrene plate",
    well_count=384,
    well_depth_mm=None,
    well_volume_ul=Unit(90.0, "microliter"),
    well_coating=None,
    sterile=False,
    is_tube=False,
    cover_types=["standard", "universal"],
    seal_types=["breathable", "ultra-clear"],
    capabilities=[
        "liquid_handle",
        "spin",
        "absorbance",
        "fluorescence",
        "luminescence",
        "incubate",
        "gel_separate",
        "gel_purify",
        "cover",
        "dispense",
        "seal",
    ],
    shortname="384-flat-white-clear",
    col_count=24,
    dead_volume_ul=Unit(7, "microliter"),
    safe_min_volume_ul=Unit(15, "microliter"),
    vendor="Corning",
    cat_no="3763",
)

#:
FALCON96UBOTTOM = ContainerType(
    name="Falcon 96-Well, Cell Culture-Treated, U-Shaped-Bottom Microplate",
    shortname="96-ubottom-clear-tc",
    is_tube=False,
    well_coating=None,
    sterile=False,
    well_count=96,
    col_count=12,
    well_depth_mm=Unit(10.5, "millimeter"),
    well_volume_ul=Unit(320, "microliter"),
    dead_volume_ul=Unit(50, "microliter"),
    safe_min_volume_ul=Unit(50, "microliter"),
    capabilities=[
        "pipette",
        "spin",
        "incubate",
        "cover",
        "stamp",
        "dispense-destination",
        "provision",
        "uncover",
        "seal",
        "unseal",
    ],
    cover_types=["universal"],
    seal_types=["ultra-clear", "foil"],
    vendor="Corning",
    cat_no="353077",
)

#:
FLAT96DELTA = ContainerType(
    name="Nunc MicroWell 96-Well, Nunclon Delta-Treated, Flat-Bottom Microplate",
    shortname="96-flat-white-dc",
    is_tube=False,
    well_coating=None,
    sterile=True,
    well_count=96,
    col_count=12,
    well_depth_mm=Unit(11.2, "millimeter"),
    well_volume_ul=Unit(320, "microliter"),
    dead_volume_ul=Unit(50, "microliter"),
    safe_min_volume_ul=Unit(50, "microliter"),
    capabilities=[
        "pipette",
        "spin",
        "incubate",
        "cover",
        "stamp",
        "dispense-destination",
        "provision",
        "uncover",
        "seal",
        "unseal",
        "liquid_handle",
        "absorbance",
        "fluorescence",
        "luminescence",
    ],
    cover_types=["low_evaporation", "standard", "universal"],
    seal_types=["ultra-clear", "foil"],
    vendor="ThermoFisher",
    cat_no="136101",
)

#:
FLAT384BLACKUBOTTOM = ContainerType(
    name="384-well Black Clear Round Bottom Ultra-low Attachment Microplate",
    shortname="384-ubottom-black-clear-tc",
    is_tube=False,
    well_coating=None,
    sterile=True,
    well_count=384,
    col_count=24,
    well_depth_mm=12.55,
    well_volume_ul=Unit(90, "microliter"),
    dead_volume_ul=Unit(10, "microliter"),
    safe_min_volume_ul=Unit(20, "microliter"),
    capabilities=[
        "pipette",
        "spin",
        "incubate",
        "cover",
        "stamp",
        "dispense",
        "dispense-destination",
        "echo_dest",
        "provision",
        "uncover",
        "seal",
        "unseal",
        "liquid_handle",
        "absorbance",
        "fluorescence",
        "luminescence",
        "bluewash",
        "spectophotometry",
    ],
    cover_types=["universal"],
    seal_types=["ultra-clear", "foil"],
    vendor="Corning",
    cat_no="3830",
)

#:
FLAT384BLACKTC = ContainerType(
    name="384-well Black Clear Flat Bottom Tissue Culture Microplate",
    shortname="384-flat-black-clear-tc",
    is_tube=False,
    well_coating=None,
    sterile=True,
    well_count=384,
    col_count=24,
    well_depth_mm=11.55,
    well_volume_ul=Unit(90, "microliter"),
    dead_volume_ul=Unit(20, "microliter"),
    safe_min_volume_ul=Unit(20, "microliter"),
    capabilities=[
        "pipette",
        "spin",
        "incubate",
        "cover",
        "stamp",
        "dispense",
        "dispense-destination",
        "echo_dest",
        "provision",
        "uncover",
        "seal",
        "unseal",
        "liquid_handle",
        "absorbance",
        "fluorescence",
        "luminescence",
        "bluewash",
        "spectophotometry",
    ],
    cover_types=["universal", "standard"],
    seal_types=["ultra-clear", "foil"],
    vendor="Corning",
    cat_no="3764",
)

_CONTAINER_TYPES = {
    "384-flat": FLAT384,
    "384-pcr": PCR384,
    "384-echo": ECHO384,
    "384-flat-white-white-lv": FLAT384WHITELV,
    "384-flat-white-white-tc": FLAT384WHITETC,
    "384-flat-clear-clear": FLAT384CLEAR,
    "96-flat": FLAT96,
    "96-flat-uv": FLAT96UV,
    "96-pcr": PCR96,
    "96-deep": DEEP96,
    "96-v-kf": V96KF,
    "96-deep-kf": DEEP96KF,
    "24-deep": DEEP24,
    "micro-2.0": MICRO2,
    "micro-1.5": MICRO15,
    "6-flat": FLAT6,
    "1-flat": FLAT1,
    "6-flat-tc": FLAT6TC,
    "res-sw96-hp": RESSW96HP,
    "res-mw8-hp": RESMW8HP,
    "res-mw12-hp": RESMW12HP,
    "96-flat-clear-clear-tc": FLAT96CLEARTC,
    "384-v-clear-clear": V384CLEAR,
    "384-round-clear-clear": ROUND384CLEAR,
    "res-sw384-lp": RESSW384LP,
    "384-echo-ldv": ECHO384LDV,
    "384-echo-ldv-plus": ECHO384LDVPLUS,
    "384-flat-white-clear": FLAT384WHITECLEAR,
    "96-well-v-bottom": V96CC,
    "96-ubottom-clear-tc": FALCON96UBOTTOM,
    "96-flat-white-dc": FLAT96DELTA,
    "384-ubottom-black-clear-tc": FLAT384BLACKUBOTTOM,
    "384-flat-black-clear-tc": FLAT384BLACKTC,
}



"""Dispense LiquidHandleMethod

Base LiquidHandleMethod used by Protocol.liquid_handle_dispense to generate a
series of movements that define a dispense from a source into a destination
"""

from ..instruction import LiquidHandle
from ..unit import Unit
from ..util import parse_unit
from .liquid_handle_method import LiquidHandleMethod


# pylint: disable=protected-access
class Dispense(LiquidHandleMethod):
    """LiquidHandleMethod for generating dispense transports

    LiquidHandleMethod for moving volume from a source well to destination wells

    Attributes
    ----------
    _liquid : LiquidClass
        used to determine parameters specific to a given type of liquid

    Notes
    -----
    The primary entry points that for this class are:
        - _aspirate_transports : generates transports from a source location
        - _prime_transports : generates transports into the priming location
                              (back into the source location or into waste)
                              for the purpose of moving the liquid-air interface
                              to the dispensing head
        - _predispense_transports : generates transports to a disposal location
        - _dispense_transports : generates transports for a destination location

    See Also
    --------
    LiquidHandleMethod : base LiquidHandleMethod with reused functionality
    """

    def __init__(self, volume_resolution=None, prime=True, predispense=True):
        super(Dispense, self).__init__()

        # parameters for required behavior
        self.volume_resolution = volume_resolution

        # parameters for optional behavior
        self.prime = prime
        self.predispense = predispense

        # LiquidHandle parameters that are generated and modified at runtime
        self._liquid = None

    def _transport_aspirate_target_volume(self, volume):
        volume_resolution = self.volume_resolution or self.default_volume_resolution(
            volume
        )

        self._transports.append(
            LiquidHandle.builders.transport(
                volume=-volume,
                mode_params=LiquidHandle.builders.mode_params(
                    volume_resolution=volume_resolution, liquid_class=self._liquid.name
                ),
            )
        )

    def _transport_dispense_target_volume(self, volume):
        volume_resolution = self.volume_resolution or self.default_volume_resolution(
            volume
        )

        self._transports.append(
            LiquidHandle.builders.transport(
                volume=volume,
                mode_params=LiquidHandle.builders.mode_params(
                    volume_resolution=volume_resolution, liquid_class=self._liquid.name
                ),
            )
        )

    def _aspirate_transports(self, volume):
        """Generates source well transports

        Generates all the transports that should happen to aspirate volume from
        the source aliquot

        Parameters
        ----------
        volume: Unit

        Returns
        -------
        list
            source well transports corresponding to the aspirate operation
        """
        self._transports = []
        volume = parse_unit(Unit(volume), "ul")
        self._transport_aspirate_target_volume(volume)
        return self._transports

    def _dispense_transports(self, volume):
        """Generates destination well transports

        Generates all the transports that should happen to dispense volume into
        the destination aliquot

        Parameters
        ----------
        volume: Unit

        Returns
        -------
        list
            destination well transports corresponding to the dispense operation
        """
        self._transports = []
        volume = parse_unit(Unit(volume), "ul")
        self._transport_dispense_target_volume(volume)
        return self._transports

    def _prime_transports(self, volume):
        """Generates transports for priming volume back into the source location

        Generates all the transports that should happen to recirculate volume
        from the source aliquot back into itself

        Parameters
        ----------
        volume: Unit

        Returns
        -------
        list
            priming transports
        """
        self._transports = []
        volume = parse_unit(Unit(volume), "ul")

        # No transports if no volume specified
        if volume == Unit("0:ul"):
            return []

        self._transport_dispense_target_volume(volume)

        return self._transports

    def _predispense_transports(self, volume):
        """Generates transports for predispensing volume back into waste

        Generates all the transports that should happen to dispense volume into
        waste before dispensing into the destination aliquots.
        This is used to clear any residual air gaps out of the dispenser.

        Parameters
        ----------
        volume: Unit

        Returns
        -------
        list
            predispense transports
        """
        self._transports = []
        volume = parse_unit(Unit(volume), "ul")

        # No transports if no volume specified
        if volume == Unit("0:ul"):
            return []

        self._transport_dispense_target_volume(volume)

        return self._transports

    # pylint: disable=unused-argument, redundant-returns-doc
    @staticmethod
    def default_volume_resolution(volume):
        """Default volume_resolution parameters

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        Unit or None
            volume_resolution

        See Also
        --------
        volume_resolution : holds any user defined volume resolution parameters
        _transport_aspirate_target_volume : generates aspirate transports
        _transport_dispense_target_volume : generates dispense transports
        """
        return None

    # pylint: disable=unused-argument
    @staticmethod
    def default_prime(volume):
        """Default priming volume

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        Unit
            prime volume
        """
        return Unit(600, "uL")

    # pylint: disable=unused-argument
    @staticmethod
    def default_predispense(volume):
        """Default predispense volume

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        Unit
            predispense volume
        """
        return Unit(10, "uL")

    def default_blowout(self, volume):
        pass

    def _calculate_overage_volume(self, volume):
        pass

    def _has_calibration(self):
        pass



"""
Contains all the Autoprotocol Informatics objects

    :copyright: 2021 by The Autoprotocol Development Team, see AUTHORS
        for more details.
    :license: BSD, see LICENSE for more details

"""

from abc import abstractmethod

from .compound import Compound
from .util import is_valid_well


class Informatics:
    """
    Base class for informatics attribute in an instruction that is to later be
    encoded as JSON.
    """

    def __init__(self):
        pass

    @abstractmethod
    def as_dict(self):
        pass

    @abstractmethod
    def validate(self):
        pass


class AttachCompounds(Informatics):
    """
    Informatics type attach_compounds is constructed as a list of dict detailing
    new compounds associated with target wells

    Parameters
    ----------
    wells: Well, list(Well) or WellGroup
        wells where compounds are associated with
    compounds: list(Compound)
        List of new compounds attached to wells
    Returns
    -------
    AttachCompounds

    """

    def __init__(self, wells, compounds):

        self.wells = wells
        self.compounds = compounds
        self.validate()
        super().__init__()

    def as_dict(self):
        """generates a Python object representation of Informatics attribute in
        class Instruction

         Returns
         -------
         dict
             a dict of python objects that have the same structure as the
             Autoprotocol JSON for the Informatics

         Notes
         -----
         Used as a part of JSON serialization of the Instruction

         See Also
         --------
         :class:`Instruction` : Instruction class

        """

        return {
            "type": "attach_compounds",
            "data": {"wells": self.wells, "compounds": self.compounds},
        }

    def validate(self):
        """
        validate input dict has valid parameters to instantiate AttachCompounds

        Raises
        ------
        TypeError
            wells is a valid well type
        TypeError
            compounds is a list
        TypeError
            compounds element is a Compound
        """
        if not is_valid_well(self.wells):
            raise TypeError(
                f"wells: {self.wells} must be Well, list of Well or WellGroup."
            )

        if not isinstance(self.compounds, list):
            raise TypeError(f"compounds: {self.compounds} must be provided in a list.")
        for compd in self.compounds:
            if not isinstance(compd, Compound):
                raise TypeError(f"compound: {compd} must be Compound type.")



"""
Module containing the harness module which helps with Manifest interpretation

    :copyright: 2021 by The Autoprotocol Development Team, see AUTHORS
        for more details.
    :license: BSD, see LICENSE for more details

"""

import argparse
import io
import json

from . import UserError
from .compound import Compound, CompoundError
from .container import WellGroup
from .protocol import Protocol
from .unit import Unit, UnitError


_DYE_TEST_RS = {"dye4000": "rs18qmhr7t9jwq", "water": "rs17gmh5wafm5p"}


def get_protocol_preview(protocol, name, manifest="manifest.json"):
    """
    Parses the 'preview' section of a manifest to use as protocol inputs

        Example Usage:

        .. code-block:: python

          p = Protocol()
          preview = get_protocol_preview(p, name="qPCR")

        Output:

        .. code-block:: python
        {
            "qPCR_input1": "value",
            "qPCR_input2": "value2",
            "qPCR_group": {
                "group_entry1": "value",
                "group_entry2": "value"
            }
        }


    Parameters
    ----------
    protocol : Protocol
        Protocol object being parsed.
    name : str
        Name of protocol in manifest to get preview for
    manifest : str, optional
        Name of manifest file

    Returns
    -------
    dict
        Dictionary of parameters used as a Protocol input

    Raises
    ------
    RuntimeError
        No manifest.json file present in directory
    RuntimeError
        Protocol not found in manifest
    RuntimeError
        More than one protocol found in manifest

    """
    try:
        manifest_json = io.open(manifest, encoding="utf-8").read()
    except IOError:
        raise RuntimeError(f"'{manifest}' file not found in directory.")
    manifest = Manifest(json.loads(manifest_json))

    source = [m for m in manifest.protocols if m["name"] == name]
    if not source:
        raise RuntimeError(
            f"Protocol '{name}' not found in list of protocols in this " f"manifest."
        )
    if len(source) != 1:
        raise RuntimeError(
            f"More than one protocol with name '{name}' was found in the "
            f"manifest. All protocol names in a manifest must be unique for it "
            f"to be valid."
        )
    preview = source[0]["preview"]
    run_params = manifest.protocol_info(name).parse(protocol, preview)

    return run_params


def param_default(type_desc):
    if isinstance(type_desc, str):
        type_desc = {"type": type_desc}

    type = type_desc["type"]  # pylint: disable=redefined-builtin
    default = type_desc.get("default")

    if default is not None and type != "group-choice":
        return default

    if type_desc["type"] in ["aliquot+", "aliquot++", "container+"]:
        return []
    elif type_desc["type"] == "group+":
        return [{}]
    elif type_desc["type"] == "group":
        return {k: param_default(v) for k, v in type_desc["inputs"].items()}
    elif type_desc["type"] == "group-choice":
        default_inputs = {}
        for option in type_desc.get("options", []):
            value = option.get("value")
            inputs = option.get("inputs")
            if inputs:
                group_typedesc = {"type": "group", "inputs": inputs}
                default_inputs[value] = param_default(group_typedesc)

        return {"value": default, "inputs": default_inputs}
    elif type_desc["type"] == "csv-table":
        return [{}, [{}]]
    else:
        return None


def convert_param(protocol, val, type_desc):
    """
    Convert parameters based on their input types

    Parameters
    ----------
    protocol : Protocol
        Protocol object being parsed.
    val : str or int or bool or dict or list
        Parameter value to be converted.
    type_desc : dict or str
        Description of input type.

    Returns
    -------
    list or dict or Unit or int or str or float or bool or None
        Converted parameter of the relevant type

    Raises
    ------
    RuntimeError
        Invalid aliquot reference provided for aliquot, aliquot+, aliquot++
    RuntimeError
        Invalid container reference provided for container, container+
    RuntimeError
        Invalid unit-type provided
    RuntimeError
        Invalid temperature condition provided
    RuntimeError
        Invalid format provided for integer or decimal
    RuntimeError
        Invalid value or format provided for group or group+
    RuntimeError
        Invalid format provided for thermocycle or thermocycle_step
    RuntimeError
        Invalid format provided for csv input
    ValueError
        Unknown input type provided

    """
    if isinstance(type_desc, str):
        type_desc = {"type": type_desc}
    if val is None:
        val = param_default(type_desc)
    if val is None:  # still None?
        return None
    type = type_desc["type"]  # pylint: disable=redefined-builtin

    if type == "aliquot":
        try:
            container = ("/").join(val.split("/")[0:-1])
            well_idx = val.split("/")[-1]
            return protocol.refs[container].container.well(well_idx)
        except (KeyError, AttributeError, ValueError):
            label = type_desc.get("label") or "[unknown]"
            raise RuntimeError(
                f"'{val}' (supplied to input '{label}') is not "
                f"a valid reference to an aliquot"
            )
    elif type == "aliquot+":
        try:
            return WellGroup([convert_param(protocol, a, "aliquot") for a in val])
        except:
            label = type_desc.get("label") or "[unknown]"
            raise RuntimeError(
                f"The value supplied to input '{label}' (type aliquot+) is "
                f"improperly formatted."
            )
    elif type == "aliquot++":
        try:
            return [convert_param(protocol, aqs, "aliquot+") for aqs in val]
        except:
            label = type_desc.get("label") or "[unknown]"
            raise RuntimeError(
                f"The value supplied to input '{label}' (type aliquot++) is "
                f"improperly formatted."
            )
    elif type == "compound":
        try:
            return Compound(val["format"], val["value"])
        except CompoundError as e:
            raise RuntimeError(f"Invalid Compound; Details: {e.value}")
    elif type == "compound+":
        try:
            return [convert_param(protocol, cont, "compound") for cont in val]
        except:
            label = type_desc.get("label") or "[unknown]"
            raise RuntimeError(
                f"The value supplied to input '{label}' (type compound+) is "
                f"improperly formatted."
            )
    elif type == "container":

        try:
            return protocol.refs[val].container
        except KeyError:
            label = type_desc.get("label") or "[unknown]"
            raise RuntimeError(
                f"'{val}' (supplied to input '{label}') is not "
                f"a valid reference to a container"
            )
    elif type == "container+":
        try:
            return [convert_param(protocol, cont, "container") for cont in val]
        except:
            label = type_desc.get("label") or "[unknown]"
            raise RuntimeError(
                f"The value supplied to input '{label}' (type container+) is "
                f"improperly formatted."
            )
    elif type in [
        "amount_concentration",
        "frequency",
        "length",
        "mass_concentration",
        "time",
        "volume",
        "volume_concentration",
        # TODO: Deprecate the following two types in next major release
        "concentration(mass)",
        "concentration(molar)",
    ]:
        try:
            return Unit(val)
        except UnitError as e:
            raise RuntimeError(
                f"The value supplied ({e.value}) as a unit of '{type}' is "
                f"improperly formatted. Units of {type} must be in the form: "
                f"'number:unit'"
            )
    elif type == "temperature":
        try:
            if val in [
                "ambient",
                "warm_30",
                "warm_35",
                "warm_37",
                "cold_4",
                "cold_20",
                "cold_80",
                "cold_196",
            ]:
                return val
            return Unit(val)
        except UnitError as e:
            raise RuntimeError(
                f"Invalid temperature value for {e.value}: "
                f"temperature input types must be either "
                f"storage conditions (ex: 'cold_20') or "
                f"temperature units in the form of "
                f"'number:unit'"
            )
    elif type in "bool":
        return bool(val)
    elif type in "csv":
        return val
    elif type in ["string", "choice"]:
        return str(val)
    elif type == "integer":
        try:
            return int(val)
        except ValueError:
            label = type_desc.get("label") or "[unknown]"
            raise RuntimeError(
                f"The value supplied to input '{label}' (type integer) is "
                f"improperly formatted."
            )
    elif type == "decimal":
        try:
            return float(val)
        except ValueError:
            label = type_desc.get("label") or "[unknown]"
            raise RuntimeError(
                f"The value supplied to input '{label}' (type decimal) is "
                f"improperly formatted."
            )
    elif type == "group":
        try:
            return {
                k: convert_param(protocol, val.get(k), type_desc["inputs"][k])
                for k in type_desc["inputs"]
            }
        except KeyError as e:
            label = type_desc.get("label") or "[unknown]"
            raise RuntimeError(
                f"The value supplied to input '{label}' (type group) is "
                f"missing a(n) {e} field."
            )
        except AttributeError:
            label = type_desc.get("label") or "[unknown]"
            raise RuntimeError(
                f"The value supplied to input '{label}' (type group) is "
                f"improperly formatted."
            )
    elif type == "group+":
        try:
            return [
                {
                    k: convert_param(protocol, x.get(k), type_desc["inputs"][k])
                    for k in type_desc["inputs"]
                }
                for x in val
            ]
        except (TypeError, AttributeError):
            raise RuntimeError(
                f"The value supplied to input '{type_desc['label']}' "
                f"(type group+) must be in the form of a list of dictionaries"
            )
        except KeyError as e:
            label = type_desc.get("label") or "[unknown]"
            raise RuntimeError(
                f"The value supplied to input '{label}' (type group+) is "
                f"missing a(n) {e} field."
            )
    elif type == "group-choice":
        try:
            return {
                "value": val["value"],
                "inputs": {
                    opt["value"]: convert_param(
                        protocol,
                        val["inputs"].get(opt["value"]),
                        {"type": "group", "inputs": opt["inputs"]},
                    )
                    for opt in type_desc["options"]
                    if opt["value"] == val["value"]
                },
            }
        except (KeyError, AttributeError) as e:
            label = type_desc.get("label") or "[unknown]"
            if e in ["value", "inputs"]:
                raise RuntimeError(
                    f"The value supplied to input '{label}' "
                    f"(type group-choice) is missing a(n) {e} field."
                )
    elif type == "thermocycle":
        try:
            return [
                {
                    "cycles": g["cycles"],
                    "steps": [
                        convert_param(protocol, s, "thermocycle_step")
                        for s in g["steps"]
                    ],
                }
                for g in val
            ]
        except (TypeError, KeyError):
            raise RuntimeError(_thermocycle_error_text())

    elif type == "thermocycle_step":
        try:
            output = {"duration": Unit(val["duration"])}
        except UnitError as e:
            raise RuntimeError(
                f"Invalid duration value for {e.value}: duration input types "
                f"must be time units in the form of 'number:unit'"
            )

        try:
            if "gradient" in val:
                output["gradient"] = {
                    "top": Unit(val["gradient"]["top"]),
                    "bottom": Unit(val["gradient"]["bottom"]),
                }
            else:
                output["temperature"] = Unit(val["temperature"])
        except UnitError as e:
            raise RuntimeError(
                f"Invalid temperature value for {e.value}: thermocycle "
                f"temperature input types must be temperature units in the "
                f"form of 'number:unit'"
            )

        if "read" in val:
            output["read"] = val["read"]

        return output

    elif type == "csv-table":
        try:
            values = []
            for i, row in enumerate(val[1]):
                value = {}
                for header, header_value in row.items():
                    type_desc = {
                        "type": val[0].get(header),
                        "label": f"csv-table item ({i}): {header}",
                    }

                    value[header] = convert_param(
                        protocol, header_value, type_desc=type_desc
                    )

                values.append(value)

            return values

        except (AttributeError, IndexError, TypeError):
            label = type_desc.get("label") or "[unknown]"
            raise RuntimeError(
                f"The values supplied to {label} (type csv-table) are "
                f"improperly formatted. Format must be a list of dictionaries "
                f"with the first dictionary comprising keys with associated "
                f"column input types."
            )

    else:
        raise ValueError(f"Unknown input type {type!r}")


class ProtocolInfo(object):
    def __init__(self, json_dict):
        self.input_types = json_dict["inputs"]

    def parse(self, protocol, inputs):
        refs = inputs["refs"]
        params = inputs["parameters"]

        for name in refs:
            ref = refs[name]
            c = protocol.ref(
                name,
                ref.get("id"),
                ref["type"],
                storage=ref.get("store"),
                discard=ref.get("discard"),
                cover=ref.get("cover"),
            )
            aqs = ref.get("aliquots")
            if aqs:
                for idx in aqs:
                    aq = aqs[idx]
                    c.well(idx).set_volume(aq["volume"])
                    if "name" in aq:
                        c.well(idx).set_name(aq["name"])
                    if "mass_mg" in aq:
                        c.well(idx).set_mass(aq["mass_mg"])
                    if "properties" in aq:
                        c.well(idx).set_properties(aq.get("properties"))

        out_params = {}
        for k in self.input_types:
            typeDesc = self.input_types[k]
            out_params[k] = convert_param(protocol, params.get(k), typeDesc)

        return out_params


class Manifest(object):
    """
    Object representation of a manifest.json file

    Parameters
    ----------
    object : JSON object
        A manifest.json file with the following format:

        .. code-block:: none

            {
              "format": "python",
              "license": "MIT",
              "description": "This is a protocol.",
              "protocols": [
                {
                  "name": "SampleProtocol",
                  "version": 1.0.0,
                  "command_string": "python sample_protocol.py",
                  "preview": {
                    "refs":{},
                    "parameters": {},
                    "inputs": {},
                    "dependencies": []
                  }
                }
              ]
            }

    """

    def __init__(self, json_dict):
        self.protocols = json_dict["protocols"]

    def protocol_info(self, name):
        try:
            return ProtocolInfo(next(p for p in self.protocols if p["name"] == name))
        except StopIteration:
            raise RuntimeError(
                f"Harness.run(): {name} does not match the "
                f"'name' field of any protocol in the "
                f"associated manifest.json file."
            )


def run(fn, protocol_name=None, seal_after_run=True, protocol_class=None):
    """
    Run the protocol specified by the function.

    If protocol_name is passed, use preview parameters from the protocol with
    the matching "name" value in the manifest.json file to run the given
    function.  Otherwise, take configuration JSON file from the command line
    and run the given function.

    Parameters
    ----------
    fn : function
        Function that generates Autoprotocol
    protocol_name :  str, optional
        str matching the "name" value in the manifest.json file
    seal_after_run : bool, optional
        Implicitly add a seal/cover to all stored refs within the protocol
        using seal_on_store()
    protocol_class: Protocol, optional
        References the base protocol class to be used for instantiation.
        If not provided, defaults to using Autoprotocol Python's default
        Protocol implementation

    Raises
    ------
    TypeError
        If protocol_class provided is not a subclass of Protocol
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("config", help="JSON-formatted protocol configuration file")
    parser.add_argument(
        "--dye_test",
        help="Execute protocol by pre-filling preview aliquots with OrangeG "
        "dye, and provisioning water only.",
        action="store_true",
    )
    args = parser.parse_args()

    source = json.loads(io.open(args.config, encoding="utf-8").read())

    if protocol_class is None:
        protocol = Protocol()
    else:
        if not issubclass(protocol_class, Protocol):
            raise TypeError(
                "Protocol class provided needs to be subclass of " "Protocol"
            )
        protocol = protocol_class()

    # pragma pylint: disable=protected-access
    if protocol_name:
        manifest_json = io.open("manifest.json", encoding="utf-8").read()
        manifest = Manifest(json.loads(manifest_json))
        params = manifest.protocol_info(protocol_name).parse(protocol, source)
        # Add dye to preview aliquots if --dye_test included as an optional
        # argument
        if args.dye_test:
            num_dye_steps = _add_dye_to_preview_refs(protocol)
    else:
        params = protocol._ref_containers_and_wells(source["parameters"])
    # pragma pylint: enable=protected-access

    try:
        fn(protocol, params)
        if seal_after_run:
            seal_on_store(protocol)
        # Convert all provisions to water if --dye_test is included as an
        # optional argument
        if args.dye_test:
            _convert_provision_instructions(
                protocol, num_dye_steps, len(protocol.instructions) - 1
            )
            _convert_dispense_instructions(
                protocol, num_dye_steps, len(protocol.instructions) - 1
            )
    except UserError as e:
        print(
            json.dumps({"errors": [{"message": e.message, "info": e.info}]}, indent=2)
        )
        return

    print(json.dumps(protocol.as_dict(), indent=2))


def _add_dye_to_preview_refs(protocol, rs=_DYE_TEST_RS["dye4000"]):
    # Store starting number of instructions
    starting_num = len(protocol.instructions)

    # For each ref in protocol
    for _, ref_obj in protocol.refs.items():

        ref_cont = ref_obj.container
        # Raise RuntimeError if any refs have an id, to avoid adding dye to
        # real samples
        if ref_cont.id:
            raise RuntimeError(
                "Cannot run a dye test when any ref has a defined container "
                "id. Please resubmit using only new containers."
            )

        # Add dye to each well
        for well in ref_cont.all_wells():
            current_vol = well.volume
            if current_vol and current_vol > Unit(0, "microliter"):
                protocol.provision(rs, well, current_vol)
                well.set_volume(current_vol)

    # Return number of instructions added
    return len(protocol.instructions) - starting_num


def _convert_provision_instructions(
    protocol, first_index, last_index, rs=_DYE_TEST_RS["water"]
):
    # Make sure inputs are valid
    if not isinstance(first_index, int):
        raise ValueError("first_index must be a non-negative integer")
    if not isinstance(last_index, int):
        raise ValueError("last_index must be a non-negative integer")
    if first_index < 0:
        raise ValueError("Indices out of range. first_index must be 0 or greater")
    if first_index > len(protocol.instructions) - 1:
        raise ValueError(
            "Indices out of range. The last instruction index in the protocol "
            "is %d" % (len(protocol.instructions) - 1)
        )
    if last_index > len(protocol.instructions) - 1:
        raise ValueError(
            "Indices out of range. The last instruction index in the protocol "
            "is %d" % (len(protocol.instructions) - 1)
        )
    if last_index < first_index:
        raise ValueError("last_index must be greater than or equal to first_index")

    for instruction in protocol.instructions[first_index : last_index + 1]:
        if instruction.op == "provision":
            instruction.data["resource_id"] = rs


def _convert_dispense_instructions(
    protocol, first_index, last_index, rs=_DYE_TEST_RS["water"]
):
    # Make sure inputs are valid
    if not isinstance(first_index, int):
        raise ValueError("first_index must be a non-negative integer")
    if not isinstance(last_index, int):
        raise ValueError("last_index must be a non-negative integer")
    if first_index < 0:
        raise ValueError("Indices out of range. first_index must be 0 or greater")
    if first_index > len(protocol.instructions) - 1:
        raise ValueError(
            "Indices out of range. The last instruction index in the protocol "
            "is %d" % (len(protocol.instructions) - 1)
        )
    if last_index > len(protocol.instructions) - 1:
        raise ValueError(
            "Indices out of range. The last instruction index in the protocol "
            "is %d" % (len(protocol.instructions) - 1)
        )
    if last_index < first_index:
        raise ValueError("last_index must be greater than or equal to first_index")

    for instruction in protocol.instructions[first_index : last_index + 1]:
        if instruction.op == "dispense":
            if "resource_id" in instruction.data:
                instruction.data["resource_id"] = rs
            if "reagent" in instruction.data:
                instruction.data.pop("reagent", None)
                instruction.data["resource_id"] = rs


def _thermocycle_error_text():
    """
    Returns formatted error text for thermocycle value errors
    """

    return """Thermocycle input types must take a list of dictionaries in the form of:
  [{"cycles": integer,
    "steps": [{
      "duration": duration,
      "temperature": temperature
      "read": boolean (optional)
    }]
  }]
--or--
  [{"cycles": integer,
    "steps": [{
      "duration": duration,
      "gradient": {
        "top": temperature,
        "bottom": temperature
      }
      "read": boolean (optional)
    }]
  }]
(You can intermix gradient and non-gradient steps)"""


def seal_on_store(protocol):
    """
    Implicitly adds seal/cover instructions to the end of a run for containers
    that do not have a cover.   Cover type applied defaults first to
    "seal" if its within the capabilities of the container type, otherwise
    to "cover".

    Example Usage:

        .. code-block:: python

            def example_method(protocol, params):
            cont = params['container']
            p.transfer(cont.well("A1"), cont.well("A2"), "10:microliter")
            p.seal(cont)
            p.unseal(cont)
            p.cover(cont)
            p.uncover(cont)

    Autoprotocol Output:

        .. code-block:: json

            {
              "refs": {
                "plate": {
                  "new": "96-pcr",
                  "store": {
                    "where": "ambient"
                  }
                }
              },
              "instructions": [
                {
                  "groups": [
                    {
                      "transfer": [
                        {
                          "volume": "10.0:microliter",
                          "to": "plate/1",
                          "from": "plate/0"
                        }
                      ]
                    }
                  ],
                  "op": "pipette"
                },
                {
                  "object": "plate",
                  "type": "ultra-clear",
                  "op": "seal"
                },
                {
                  "object": "plate",
                  "op": "unseal"
                },
                {
                  "lid": "universal",
                  "object": "plate",
                  "op": "cover"
                },
                {
                  "object": "plate",
                  "op": "uncover"
                },
                {
                  "type": "ultra-clear",
                  "object": "plate",
                  "op": "seal"
                }
              ]
            }

    """
    for _, ref in protocol.refs.items():
        if "store" in ref.opts.keys():
            if not (ref.container.is_covered() or ref.container.is_sealed()):
                default_method = ref.container.container_type.prioritize_seal_or_cover
                sealable = "seal" in ref.container.container_type.capabilities
                coverable = "cover" in ref.container.container_type.capabilities
                if default_method == "seal" and sealable:
                    protocol.seal(
                        ref.container, ref.container.container_type.seal_types[0]
                    )
                elif default_method == "cover" and coverable:
                    protocol.cover(
                        ref.container, ref.container.container_type.cover_types[0]
                    )
                elif sealable:
                    protocol.seal(
                        ref.container, ref.container.container_type.seal_types[0]
                    )
                elif coverable:
                    protocol.cover(
                        ref.container, ref.container.container_type.cover_types[0]
                    )
                else:
                    continue
